@@ -2,12 +2,64 @@
 #define GFF_UTILS_H
 #include "gff.h"
 #include "GStr.h"
+#include "GVec.hh"
 #include "GFaSeqGet.h"
 
 extern bool verbose;
 extern bool debugMode;
+extern bool ensembl_convert;
 
-typedef bool GFValidateFunc(GffObj* gf, GList<GffObj>* gfadd);
+extern FILE* ffasta;
+extern FILE* f_in;
+extern FILE* f_out;
+extern FILE* f_w; //writing fasta with spliced exons (transcripts)
+extern int wPadding; //padding for -w option
+extern FILE* f_x; //writing fasta with spliced CDS
+extern FILE* f_y; //wrting fasta with translated CDS
+
+extern int maxintron;
+
+extern bool wCDSonly;
+extern bool wNConly;
+extern int minLen; //minimum transcript length
+extern bool validCDSonly; // translation with no in-frame STOP
+extern bool bothStrands; //for single-exon mRNA validation, check the other strand too
+extern bool altPhases; //if original phase fails translation validation,
+                     //try the other 2 phases until one makes it
+extern bool addCDSattrs;
+extern bool add_hasCDS;
+
+extern bool adjustStop; //automatic adjust the CDS stop coordinate
+extern bool covInfo; // --cov-info : only report genome coverage
+extern GStr tableFormat; //list of "attributes" to print in tab delimited format
+extern bool spliceCheck; //only known splice-sites
+extern bool decodeChars; //decode url-encoded chars in attrs (-D)
+extern bool StarStop; //use * instead of . for stop codon translation
+extern bool fullCDSonly; // starts with START, ends with STOP codon
+
+extern bool multiExon;
+extern bool writeExonSegs;
+extern char* tracklabel;
+extern char* rfltGSeq;
+extern char rfltStrand;
+extern uint rfltStart;
+extern uint rfltEnd;
+extern bool rfltWithin; //check for full containment within given range
+extern bool addDescr;
+
+
+extern bool fmtGFF3; //output: GFF3
+//other formats only make sense in transcriptOnly mode
+extern bool fmtGTF;
+extern bool fmtBED;
+extern bool fmtTLF;
+extern bool fmtTable;
+
+
+extern GffPrintMode exonPrinting;
+
+//typedef bool GFValidateFunc(GffObj* gf, GList<GffObj>* gfadd);
+typedef bool GFValidateFunc(GffObj* gf);
 
 //test if a transcript should be printed (and not printed yet)
 #define T_PRINTABLE(d) (((d) & 0x100)==0)
@@ -26,7 +78,90 @@ typedef bool GFValidateFunc(GffObj* gf, GList<GffObj>* gfadd);
 //keep/set original/old strand
 #define T_SET_OSTRAND(d, s) d |= s
 
-class GeneInfo { //for Ensembl GTF conversion
+class SeqInfo { //populated from the -s option of gffread
+ public:
+  int len;
+  char* descr;
+  SeqInfo( int l, char* s): len(l), descr(NULL) {
+    if (s!=NULL)
+      descr=Gstrdup(s);
+  }
+  ~SeqInfo() {
+    GFREE(descr);
+  }
+};
+
+class RefTran {
+ public:
+   char* new_name;
+   RefTran(char *ns) {
+      new_name=NULL;
+      if (ns!=NULL)
+         new_name=Gstrdup(ns);
+      }
+   ~RefTran() {
+      GFREE(new_name);
+      }
+};
+
+extern GFastaDb gfasta;
+extern GHash<SeqInfo> seqinfo;
+extern GHash<int> isoCounter; //counts the valid isoforms
+extern GHash<RefTran> reftbl;
+
+char* getSeqDescr(char* seqid);
+char* getSeqName(char* seqid);
+int adjust_stopcodon(GffObj& gffrec, int adj, GList<GSeg>* seglst=NULL);
+void printTableData(FILE* f, GffObj& g, bool inFasta=false);
+bool validateGffRec(GffObj* gffrec);
+bool process_transcript(GFastaDb& gfasta, GffObj& gffrec);
+
+enum ETableFieldType {
+  ctfGFF_Attr=0, // attribute name as is
+  ctfGFF_ID, //ID or @id or transcript_id
+  ctfGFF_geneID, //geneID or @gene_id or @geneid
+  ctfGFF_geneName, //geneName or @gene_name or @genename
+  ctfGFF_Parent, //Parent or @parent
+  ctfGFF_chr, //@chr
+  ctfGFF_feature, //@feature
+  ctfGFF_start, //@start
+  ctfGFF_end, //@end
+  ctfGFF_strand, //@strand
+  ctfGFF_numexons, //@numexons
+  ctfGFF_exons, //@exons
+  ctfGFF_cds, //@cds
+  ctfGFF_covlen, //@covlen
+  ctfGFF_cdslen//@cdslen
+};
+
+class CTableField {
+ public:
+   ETableFieldType type;
+   GStr name; //only for type ctfGFF_Attr
+   CTableField(ETableFieldType atype=ctfGFF_Attr):type(atype) { }
+   CTableField(GStr& attrname):type(ctfGFF_Attr),name(attrname) { }
+};
+
+
+extern GVec<CTableField> tableCols; //table output format fields
+
+class GffLocus;
+class GenomicSeqData;
+class GeneInfo;
+
+class GTData { // transcript associated data
+ public:
+   GffObj* rna;
+   GenomicSeqData* gdata;
+   GffLocus* locus;
+   GffObj* replaced_by;
+   GeneInfo* geneinfo;
+   GTData(GffObj* t=NULL, GenomicSeqData* gd=NULL);
+   bool operator<(GTData& b) { return (rna < b.rna); }
+   bool operator==(GTData& b) { return (rna==b.rna); }
+};
+
+class GeneInfo {
  public:
    int flag;
    GffObj* gf;
@@ -36,17 +171,23 @@ class GeneInfo { //for Ensembl GTF conversion
      gf=NULL;
      flag=0;
    }
-   GeneInfo(GffObj* gfrec, bool ensembl_convert=false):gene_names(true, true, true),
+
+   GeneInfo(GffObj* gfrec, GenomicSeqData* gdata, bool ensembl_convert=false):flag(0), gf(NULL), gene_names(true, true, true),
                     transcripts(true,true,true) {
-     flag=0;
      if (gfrec->getGeneName())
         gene_names.Add(new GStr(gfrec->getGeneName()));
      transcripts.Add(new GStr(gfrec->getID()));
-     create_gf(gfrec, ensembl_convert);
-     }
+     create_gf(gfrec, gdata ,ensembl_convert);
+   }
 
-   void create_gf(GffObj* gfrec, bool ensembl_convert) {
+   ~GeneInfo() {
+      delete gf;
+   }
+
+   void create_gf(GffObj* gfrec, GenomicSeqData* gdata, bool ensembl_convert) {
      gf=new GffObj(gfrec->getGeneID());
+     GTData* gfdata=new GTData(gf, gdata);
+     gfdata->geneinfo=this;
      gf->gseq_id=gfrec->gseq_id;
      gf->track_id=gfrec->track_id;
      gf->start=gfrec->start;
@@ -55,28 +196,30 @@ class GeneInfo { //for Ensembl GTF conversion
      gf->setFeatureName("gene");
      gf->isGene(true);
      gf->isUsed(true);
-     gf->uptr=this;
+     //gf->uptr=gfdata; //for these new gene objects
      gfrec->incLevel();
      gfrec->parent=gf;
      gf->children.Add(gfrec);
+     const char* s=NULL;
+     if ((s=gfrec->getGeneName())) {
+    	 gf->addAttr("Name", s);
+    	 gf->copyAttrs(gfrec);
+     }
      if (ensembl_convert) {
        //gf->addAttr("type", gf->getTrackName());
        const char* biotype=gfrec->getAttr("type");
        if (biotype) gf->addAttr("type", biotype);
        }
-     //gf->children.Add(gfrec);
-     }
-   //~GeneInfo() {
-   //  }
+       // gf->children.Add(gfrec);
+   }
+
    void update(GffObj* gfrec) {
      if (transcripts.AddedIfNew(new GStr(gfrec->getID()))<0)
        return;
      gene_names.AddedIfNew(new GStr(gfrec->getGeneName()));
      if (gf==NULL) {
         GError("GeneInfo::update() called on uninitialized gf!\n");
-        //create_gf(gfrec);
-        //return;
-        }
+     }
      gfrec->parent=gf;
      gf->children.Add(gfrec);
      gfrec->incLevel();
@@ -85,53 +228,49 @@ class GeneInfo { //for Ensembl GTF conversion
      if (gf->end<gfrec->end)
            gf->end=gfrec->end;
      }
-    void finalize() {
+
+   void finalize() {
      //prepare attributes for printing
      //must be called right before printing
      if (gf==NULL || transcripts.Count()==0) return;
      if (gene_names.Count()>0) {
        gf->addAttr("Name", gene_names[0]->chars());
-       /*
-       GStr s(gene_names[0]->chars());
-       for (int i=1;i<gene_names.Count();i++) {
-          s.append(",");
-          s.append(gene_names[i]->chars());
-          }
-       gf->addAttr("genes", s.chars());
-       */
-       } //has gene names
-       GStr t(transcripts[0]->chars());
-       for (int i=1;i<transcripts.Count();i++) {
+     } //has gene names
+     GStr t(transcripts[0]->chars());
+     for (int i=1;i<transcripts.Count();i++) {
           t.append(",");
           t.append(transcripts[i]->chars());
-          }
-       gf->addAttr("transcripts", t.chars());
      }
-};
-
-
-class GffLocus;
-
-class GTData { //transcript associated data
- public:
-   GffObj* rna;
-   GffLocus* locus;
-   GffObj* replaced_by;
-   GeneInfo* geneinfo;
-   //int flag;
-   GTData(GffObj* t=NULL) {
-       rna=t;
-       //flag=0;
-       locus=NULL;
-       replaced_by=NULL;
-       geneinfo=NULL;
-       if (rna!=NULL) {
-           geneinfo=(GeneInfo*)rna->uptr; //take over geneinfo, if there
-           rna->uptr=this;
-       }
+     gf->addAttr("transcripts", t.chars());
    }
-   bool operator<(GTData& b) { return (rna < b.rna); }
-   bool operator==(GTData& b) { return (rna==b.rna); }
+};
+class GenomicSeqData {
+  int gseq_id;
+ public:
+  const char* gseq_name;
+  int seqreg_start; //if given by ##sequence-region comment
+  int seqreg_end;
+  GList<GffObj> gfs; //all non-transcript features -> usually gene features
+  GList<GffObj> rnas; //all transcripts on this genomic sequence
+  GList<GffLocus> loci; //all loci clusters
+  GList<GTData> tdata; //transcript data (uptr holder for all rnas loaded here)
+  uint64 f_bases;//base coverage on forward strand
+  uint64 r_bases;//base coverage on reverse strand
+  uint64 u_bases;//base coverage on undetermined strand
+  //GenomicSeqData(int gid=-1):rnas(true,true,false),loci(true,true,true),
+  GenomicSeqData(int gid=-1):gseq_id(gid), gseq_name(NULL), seqreg_start(0), seqreg_end(0),
+		  gfs(true, true, false),rnas((GCompareProc*)gfo_cmpByLoc),loci(true,true,false),
+		  tdata(false,true,false),  f_bases(0), r_bases(0), u_bases(0) {
+  if (gseq_id>=0)
+    gseq_name=GffObj::names->gseqs.getName(gseq_id);
+  }
+
+  bool operator==(GenomicSeqData& d){
+    return gseq_id==d.gseq_id;
+  }
+  bool operator<(GenomicSeqData& d){
+    return (gseq_id<d.gseq_id);
+  }
 };
 
 class CGeneSym {
@@ -450,34 +589,6 @@ public:
     }
 };
 
-class GenomicSeqData {
-  int gseq_id;
- public:
-  const char* gseq_name;
-  int seqreg_start; //if given by ##sequence-region comment
-  int seqreg_end;
-  GList<GffObj> gfs; //all non-transcript features -> usually gene features
-  GList<GffObj> rnas; //all transcripts on this genomic sequence
-  GList<GffLocus> loci; //all loci clusters
-  GList<GTData> tdata; //transcript data (uptr holder for all rnas loaded here)
-  uint64 f_bases;//base coverage on forward strand
-  uint64 r_bases;//base coverage on reverse strand
-  uint64 u_bases;//base coverage on undetermined strand
-  //GenomicSeqData(int gid=-1):rnas(true,true,false),loci(true,true,true),
-  GenomicSeqData(int gid=-1):gseq_id(gid), gseq_name(NULL), seqreg_start(0), seqreg_end(0),
-		  gfs(true, true, false),rnas((GCompareProc*)gfo_cmpByLoc),loci(true,true,false),
-		  tdata(false,true,false),  f_bases(0), r_bases(0), u_bases(0) {
-  if (gseq_id>=0)
-    gseq_name=GffObj::names->gseqs.getName(gseq_id);
-  }
-
-  bool operator==(GenomicSeqData& d){
-    return gseq_id==d.gseq_id;
-  }
-  bool operator<(GenomicSeqData& d){
-    return (gseq_id<d.gseq_id);
-  }
-};
 
 int gseqCmpName(const pointer p1, const pointer p2);
 
@@ -574,6 +685,7 @@ class GffLoader {
 		bool fuzzSpan:1; //matching/contained redundancy relaxed to disregard full boundary containment
 		bool dOvlSET:1; //discard overlapping Single Exon Transcripts on any strand
 		bool forceExons:1;
+		bool streamIn:1;
 	  };
   };
 
@@ -599,7 +711,9 @@ class GffLoader {
       }
   }
 
-  void load(GList<GenomicSeqData>&seqdata, GFValidateFunc* gf_validate=NULL, GFFCommentParser* gf_parsecomment=NULL);
+  bool validateGffRec(GffObj* gffrec);
+
+  void load(GList<GenomicSeqData>&seqdata, GFFCommentParser* gf_parsecomment=NULL);
 
   bool placeGf(GffObj* t, GenomicSeqData* gdata);
 
